import asyncio
import inspect
from functools import wraps


def task(afunc, *exogenous_deps):
    """
    Example:

        @task
        async def A():
            pass

        @task
        async def B(val=None):
            pass

        async def main():
            # example 1
            a = A()
            b = B(a)

            # example 2
            a = A()
            # b depends on a, but dependency is explicitly stated in task wrapper, not through `val` param
            b = task(B, a)()

    :param afunc: function that creates an awaitable on invocation
    :param exogenous_deps: awaitable
    :return:
    """
    @wraps(afunc)
    def wrapper(*args, **kwargs):
        async def _inner():
            callargs = inspect.signature(afunc).bind(*args, **kwargs).arguments

            gather_args = {}
            non_gather_args = {}
            for k, v in callargs.items():
                if inspect.isawaitable(v):
                    gather_args[k] = v
                else:
                    non_gather_args[k] = v

            gather_args = dict(
                zip(
                    gather_args.keys(),
                    await asyncio.gather(*gather_args.values())
                )
            )
            callargs = {**gather_args, **non_gather_args}

            await asyncio.gather(*exogenous_deps)

            return await afunc(**callargs)
        return asyncio.create_task(_inner())
    return wrapper



import sys
from setuptools import setup, find_packages


if sys.version_info < (3, 7):
    raise RuntimeError('aiodag requires Python 3.7+')


classifiers = [
    'License :: OSI Approved :: MIT License',
    'Intended Audience :: Developers',
    'Programming Language :: Python :: 3',
    'Framework :: AsyncIO',
]


keywords = ["dag", "asyncio", "aiodag"]


setup(
    name='aiodag',
    version='0.3',
    description=('Build and execute AsyncIO powered DAGs.'),
    classifiers=classifiers,
    platforms=['POSIX'],
    author="aa1371",
    url='https://github.com/aa1371/aiodag',
    download_url='https://github.com/aa1371/aiodag/archive/refs/tags/v0.3.tar.gz',
    license='MIT',
    packages=find_packages(exclude=['tests', 'tests.*']),
    keywords=keywords
)
